/*
  Author : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : 28/03/2001

  Goals : writing a replacement of libusb, to read/write over endpoint
  in a more convenient way.

*/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <signal.h>
#include <errno.h>

#include <linux/usb.h>
#include <asm/page.h>

#include "usbfs.h"

/* for ident(1) command */
static char id[] = "@(#) $Id: usbfs.c,v 1.10 2001/04/08 22:31:21 papillau Exp $";

/* Device descriptor */
struct usb_device_descriptor {
	__u8  bLength;
	__u8  bDescriptorType;
	__u16 bcdUSB;
	__u8  bDeviceClass;
	__u8  bDeviceSubClass;
	__u8  bDeviceProtocol;
	__u8  bMaxPacketSize0;
	__u16 idVendor;
	__u16 idProduct;
	__u16 bcdDevice;
	__u8  iManufacturer;
	__u8  iProduct;
	__u8  iSerialNumber;
	__u8  bNumConfigurations;
} __attribute__ ((packed));

/* try to open the file and get USB device information,
   if it's ok, check if it matches vendorID & productID and return
   a usable file descriptor in this case. Else returns -1.
*/

int test_file(const char *path, int vendorID, int productID)
{
  int fd;
  struct usb_device_descriptor desc;
  
  fd = open(path,O_RDWR);
  if (fd == -1)
    {
      perror(path);
      return -1;
    }
  
  if (read(fd,&desc,sizeof(desc)) == sizeof(desc))
    {
      /* great, we read something */
      
      /* check, it match the correct structure */
      
      if (desc.bLength == sizeof(desc))
	{
	  /*	  
	  printf("=== %s ===\n",path);
	  printf("  bLength         = %u\n",desc.bLength);
	  printf("  bDescriptorType = %u\n",desc.bDescriptorType);
	  printf("  bcdUSB          = %04x\n",desc.bcdUSB);
	  printf("  idVendor        = %04x\n",desc.idVendor);
	  printf("  idProduct       = %04x\n",desc.idProduct);
	  printf("  bcdDevice       = %04x\n",desc.bcdDevice);
	  */
	  if (vendorID == desc.idVendor
	      && productID == desc.idProduct)
	    return fd;
	}
    }
  
  close(fd);
  return -1;
}

/* search for a vendorID, productID. return -1 if search failed
   or a usable file descriptor
*/

int usbfs_search(const char *path, int vendorID, int productID)
{
  int result = -1;
  DIR * dir;
  struct dirent * dirp;
  
  dir = opendir(path);
  if (dir == NULL)
    {
      perror(path);
      return -1;
    }
  
  while ((dirp=readdir(dir)) != NULL)
    {
      char file[PATH_MAX+1];
      struct stat statbuf;
      
      if (strcmp(dirp->d_name,".")==0)
	continue;
      if (strcmp(dirp->d_name,"..")==0)
	continue;
      
      sprintf(file,"%s/%s",path,dirp->d_name);
      
      if (stat(file,&statbuf)!=0)
	{
	  perror(file);
	  continue;
	}
      
      if (S_ISDIR(statbuf.st_mode))
	{
	  if ((result = usbfs_search(file,vendorID,productID)) < 0)
	    continue;
	  else
	    break;
	}
      
      if (S_ISREG(statbuf.st_mode))
	{
	  if ((result=test_file(file,vendorID,productID)) < 0)
	    continue;
	  else
	    break;
	}
    }

  closedir(dir);
  return result;
}

/* ====================== START OF LIBRARY ==================================*/

/* USBDEVFS_CONTROL => proc_control() => usb_control_msg() */

int usbfs_control_msg(int fd,int requesttype,int request,
		      int value,int idx,void * buf,int len, int timeout)
{
  int ret;
  struct usbdevfs_ctrltransfer ctrl;

  ctrl.requesttype = requesttype;
  ctrl.request     = request;
  ctrl.value       = value;
  ctrl.index       = idx;
  ctrl.length      = len;
  ctrl.timeout     = timeout;
  ctrl.data        = buf;

  ret = ioctl(fd,USBDEVFS_CONTROL,&ctrl);
  return ret;
}

/* USBDEVFS_BULK => proc_bulk() => usb_bulk_msg() */

/* from "devio.c", bulk.len <= PAGE_SIZE, else -EINVAL is returned.
   else the real length read is returned (see proc_bulk()).

   myusb_bulk_msg() reads/write what is available. timeout is specified in
   ms. the number of bytes read/written is returned or -1 if an error occurs.
*/

int usbfs_bulk_msg(int fd,int ep,unsigned char *buf,int size, int timeout)
{
  struct usbdevfs_bulktransfer bulk;
  int ret, received = 0;

  do
    {
      bulk.ep      = ep;
      bulk.len     = size;
      if (bulk.len > PAGE_SIZE)
	bulk.len = PAGE_SIZE;
      bulk.timeout = timeout;
      bulk.data    = buf;
      
      ret = ioctl(fd,USBDEVFS_BULK,&bulk);
      if (ret < 0)
	return ret;

      buf  += ret;
      size -= ret;

      received += ret;
    }
  while (ret==bulk.len && size>0);

  return received;
}

/* USBDEVFS_RESETEP => proc_resetep() => usb_settoggle() */ 

int usbfs_resetep(int fd,int ep)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_RESETEP,&ep);
  return ret;
}

/*
  USBDEVFS_RESET
*/

/*  USBDEVFS_CLEAR_HALT => proc_clearhalt => usb_clear_halt */

int usbfs_clear_halt(int fd, int ep)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_CLEAR_HALT,&ep);
  return ret;
}

/* USBDEVFS_GETDRIVER => proc_getdriver() => nothing */

/* USBDEVFS_CONNECTINFO => proc_connectinfo() => nothing */

/* USBDEVFS_SETINTERFACE => proc_setintf() => usb_set_interface() */

int usbfs_set_interface(int fd, int interface, int altsetting)
{
  struct usbdevfs_setinterface setintf;
  int ret;

  setintf.interface = interface;
  setintf.altsetting = altsetting;

  ret = ioctl(fd,USBDEVFS_SETINTERFACE,&setintf);
  return ret;
}

/* USBDEVFS_SETCONFIGURATION => proc_setconfig => usb_set_configuration */

int usbfs_set_configuration(int fd,int config)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_SETCONFIGURATION,&config);
  return ret;
}

/* USBDEVFS_SUBMITURB => proc_submiturb() => usb_submit_urb() */

int usbfs_submit_urb(int fd, struct usbdevfs_urb *urb)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_SUBMITURB,urb);
  return ret;
}

/* USBDEVFS_DISCARDURB */

/* USBDEVFS_REAPURB => proc_reapurb() => nothing.

   On return urb points to a pointer of type struct usbdevfs_urb.
   This pointer is the same one, passed to myusb_submit_urb()
*/

int usbfs_reapurb(int fd,struct usbdevfs_urb **purb)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_REAPURB,purb);
  return ret;
}

/*  USBDEVFS_REAPURBNDELAY */

int usbfs_reapurb_ndelay(int fd,struct usbdevfs_urb **purb)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_REAPURBNDELAY,purb);
  return ret;
}

/*
  USVDEVFS_DISCSIGNAL
*/

/* USBDEVFS_CLAIMINTERFACE => proc_claiminterface => claimintf() */

int usbfs_claimintf(int fd, int interface)
{
  int ret;

  ret = ioctl(fd, USBDEVFS_CLAIMINTERFACE,&interface);
  return ret;
}

/* USBDEVFS_RELEASEINTERFACE =>  proc_releaseinterface => releaseintf */

int usbfs_releaseintf(int fd, int interface)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_RELEASEINTERFACE,&interface);
  return ret;
}

/* USBDEVFS_IOCTL => proc_ioctl() => driver->ioctl() */

int usbfs_ioctl(int fd,struct usbdevfs_ioctl * ctrl)
{
  int ret;

  ret = ioctl(fd,USBDEVFS_IOCTL,ctrl);
  return ret;
}

/* ======================= END OF LIBRARY ===================================*/

/* === BEGIN OF UTILITY === */

int USB_read_1(int fd,void * buf, int size)
{
  int ret;
  struct timeval debut, fin;

  gettimeofday(&debut,NULL);
  ret = usbfs_bulk_msg(fd,0x85,buf,size,50);
  gettimeofday(&fin,NULL);

  printf("USB_read_1 (%d) = %d in %f ms\n",size,ret,
	 (double)(fin.tv_sec-debut.tv_sec)*1000.0
	 +(double)(fin.tv_usec-debut.tv_usec)/1000.0);

  return ret;
}

void usbfs_fill_bulk_urb(struct usbdevfs_urb * urb,
			 int ep, void *buf, size_t count, int sig)
{
  memset(urb,0,sizeof(*urb));
  /*
    unsigned char type;
    unsigned char endpoint;
    int status;
    unsigned int flags;
    void *buffer;
    int buffer_length;
    int actual_length;
    int start_frame;
    int number_of_packets;
    int error_count;
    unsigned int signr;
    void *usercontext;
    struct usbdevfs_iso_packet_desc iso_frame_desc[0];
  */

  urb->type = USBDEVFS_URB_TYPE_BULK;
  urb->endpoint = ep;
  urb->flags = USBDEVFS_URB_QUEUE_BULK; /* set this to queue several URB,
					   see usb-uhci:1635, or you will get
					   a ENXIO error from usbfs_submit_urb
					*/
  urb->buffer = buf;
  urb->buffer_length = count;
  urb->signr = sig; /*SIGRTMIN;*/

  /* signr : 0 for no signal, or a signal between SIGRTMIN & SIGRTMAX */
}

int USB_read_small(int fd,int ep, void * buf, size_t count)
{
  int ret;
  struct usbdevfs_urb my_urb, *purb;

  if (count > PAGE_SIZE)
    count = PAGE_SIZE;

  usbfs_fill_bulk_urb(&my_urb,ep,buf,count,0);

  ret = usbfs_submit_urb(fd,&my_urb);
  if (ret < 0)
    perror("usbfs_submit_urb");
  else
    {
      ret = usbfs_reapurb(fd,&purb);
      if (ret < 0)
	perror("usbfs_reapurb");
      else
	{
	  ret = my_urb.actual_length;

	  if (my_urb.status == -EPIPE)
	    if (usbfs_clear_halt(fd,ep) < 0)
	      {
		perror("usbfs_clear_halt");
		fflush(stderr);
	      }
	}
    }
  return ret;
}

int usbfs_bulk(int fd,int ep, void *buf, size_t count)
{
  int ret, received = 0, small_count;

  do
    {
      ret = USB_read_small(fd,ep,buf,count);
      if (ret < 0)
	return ret;

      buf   += ret;
      count -= ret;

      received += ret;
    }
  while (ret == small_count && count>0);

  return received;
}

/* usbfs_new_bulk : basic design is to have several (at least 2) URB active
   all the time and return when one of them as completed. We will use
   static buffers & static variables to do so.
*/

#define NB_URB 10 /* should be enought */
#define URB_BUF_SIZE 53 /* OK : 53 is to make cell_read() happy,
			   it will be changed in the future */

int usbfs_new_bulk(int fd, int ep, void *buf, size_t count)
{
  static unsigned char buffer[NB_URB][URB_BUF_SIZE];
  static struct usbdevfs_urb urb[NB_URB];
  static int urb_submitted[NB_URB];
  static int first_time = 1;

  int i, ret;
  struct usbdevfs_urb * purb;

  printf("usbfs_new_bulk\n");
  fflush(stdout);

  /* first time initialization : will be done only once */
  if (first_time)
    {
      first_time = 0;
      for (i=0;i<NB_URB;i++)
	urb_submitted[i] = 0;
    }

  /* submit all URB which have not been submitted */

  for (i=0;i<NB_URB;i++)
    if (!urb_submitted[i])
      {
	/* to be sure we does not get previous data : zeroed the buffer */
	memset(buffer[i],0,sizeof(buffer[i]));
	usbfs_fill_bulk_urb(&urb[i],ep,buffer[i],sizeof(buffer[i]),0);

	if ((ret=usbfs_submit_urb(fd,&urb[i])) < 0)
	  {
	    perror("oh no! usbfs_submit_urb");
	    return ret;
	  }

	printf("URB %d submitted\n",i);
	fflush(stdout);
	urb_submitted[i] = 1;

	/* rest a little bit :-) */

	usleep(1000);
      }

  /* wait that at least one URB completed */

  if ((ret=usbfs_reapurb(fd,&purb)) < 0)
    {
      perror("usbfs_reapurb");
      return ret;
    }

  /* just for debuggin, we search which URB has completed */

  for (i=0;i<NB_URB;i++)
    if (purb == &urb[i])
      {
	printf("URB %d completed\n",i);

	/* ... and mark it as not submitted again */
	urb_submitted[i] = 0;

	/* stop the search */
	break;
      }

  /* hey, we got data! copy to the user buffer */

  printf("status = %d\n",purb->status);
  printf("error_count = %d\n",purb->error_count);
  printf("actual_length = %d\n",purb->actual_length);
  fflush(stdout);

  if (purb->actual_length > count)
    {
      printf("argh: user buffer too small, discarding data!\n");
      fflush(stdout);
      return -1;
    }

  memcpy(buf,purb->buffer,purb->actual_length);
  return purb->actual_length;
}
